"""

    LDAPExceptions.py
    $Author: stevray $
    $Date: 2002/02/13 18:24:06 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    Defines the LDAP exceptions that STLDAP traps and/or raises.

    9/28/01 Don Stevens-Rayburn

        Added the scode attribute to each exception.  This attribute gives the
        ActiveX Exception Number for the error. This allows us to use the
        exception class to raise a COM exception in the Windows environment
        without changing the effectiveness in the pure Python environment.

        To raise a python exception, simply use the code

            raise NO_SUCH_ATTRIBUTE(source = '...', information = '...' )

        To raise the equivalent COM exception, use the code

            from win32com.server.exception import COMException

            raise COMExeption(description=NO_SUCH_ATTRIBUTE.__doc__,
                              source='...',
                              scode=NO_SUCH_ATTRIBUTE.scode)
        

"""    

__version__ = '$Revision: 1.8 $'[11:-2]

import exceptions

class LDAPException(exceptions.Exception):
    """
        Base class for all exceptions that the STLDAP package
        raises.
    """
    scode = 0x8005a001
    def __init__(self, source = None, information = None):
        self.source = source
        self.information = information

    def __str__(self):
        msg = '\n   Description:     ' + self.__doc__
        if self.source:
            msg = msg + '\n   Source: %s' % (self.source.__class__)
            for item in self.source.__dict__.keys():
                msg = msg + '\n   %s: %s ' % (item, self.source.__dict__[item])
        if self.information:
            msg = msg + '\n   Additional Info: ' + self.information
        return msg

    def __repr__(self):
        return self.__str__()

class ADD_ATTRIBUTE_FAILURE(LDAPException):
    """ Failed to add an attribute or and attribute value. """
    scode = 0x8005a00d

class ALREADY_EXISTS(LDAPException):
    """ Target object already exists. """
    scode = 0x8005a016

class BIND_FAILURE(LDAPException):
    """ Failed to perform an anonymous bind to the LDAP server. """
    scode = 0x8005a00b

class DELETE_FAILURE(LDAPException):
    """ Failed to delete the requested object. """
    scode = 0x8005a00c

class INSUFFICIENT_ACCESS(LDAPException):
    """ Insufficient access to perform operation """
    scode = 0x8005a007

class INVALID_CREDENTIALS(LDAPException):
    """ User credentials are invalid. """
    scode = 0x8005a009

class INVALID_DN_SYNTAX(LDAPException):
    """ Invalid distinguished name syntax. """
    scode = 0x8005a015

class MAIL_ACCOUNT_EXISTS(LDAPException):
    """ User already is a mail recipient. """
    scode = 0x8005a006

class NO_GROUP_NAME(LDAPException):
    """ Failed to find a group name. """
    scode = 0x8005a004

class NO_OBJECTCLASSES_SPECIFIED(LDAPException):
    """ An objectClasses argument must be specified. """
    scode = 0x8055a013

class NO_OU_NAME(LDAPException):
    """ Failed to find an organizational unit name. """
    scode = 0x8005a005

class NO_SUCH_OBJECT(LDAPException):
    """ No such object """
    scode = 0x8005a001

class NO_SUCH_ATTRIBUTE(LDAPException):
    """ No such attribute """
    scode = 0x8005a002

class NO_SUCH_VALUE(LDAPException):
    """ No such value """

class NOT_ALLOWED_ON_NONLEAF(LDAPException):
    """ The requested operation cannot be performed on a non-leaf. """
    scode = 0x8005a00f

class NOT_ENOUGH_INFORMATION(LDAPException):
    """ You must give either a distinguished name or both the parent dn and ou or group name. """
    scode = 0x8005a003

class OBJECT_ALREADY_EXISTS(LDAPException):
    """ The target object already exists. """
    scode = 0x8005a010

class OPEN_FAILURE(LDAPException):
    """ Failed to open connection to the LDAP server. """
    scode = 0x8005a00a

class RENAME_ERROR(LDAPException):
    """ Rename operation failed. """
    scode = 0x8005a014

class REQUIRED_ATTRIBUTE(LDAPException):
    """ Attribute is required, you cannot delete it. """
    scode = 0x8005a012

class UNRECOGNIZED_OBJECT(LDAPException):
    """ Object is not one that I can deal with. """
    scode = 0x8005a008
    
class USER_ALREADY_EXISTS(LDAPException):
    """ This user already exists. """
    scode = 0x8005a00e

class USER_MOVE_FAILURE(LDAPException):
    """ Failed to move the user. """
    scode = 0x8005a011

