"""

    STLDAPManager.py
    $Author: payne $
    $Date: 2002/08/06 15:50:52 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

ST LDAP Manager class.

A class that lets users authenticate against an LDAP directory, and to
search, browse, and modify the directory to an extent that depends on
their roles.

"""
__version__ = '$Revision: 1.27 $'[11:-2]

from Connection import AnonymousUser, AuthenticatedUser
from LDAPDirectory import LDAPDirectory
from LDAPExceptions import *
from LDAPUtilities import _ldap
from ZObject import ZObject
from ZSchema import ZSchema

from AccessControl import ClassSecurityInfo, getSecurityManager
from Globals import DTMLFile, InitializeClass
from OFS.SimpleItem import SimpleItem

class STLDAPManager( SimpleItem ):
    """
    A persistent object for managing objects in an LDAP directory.  
    Anonymous users may search, and authenticated users may update 
    the directory, limited by their roles.
    """

    meta_type = "LDAP Directory Manager"

    manage_options = (
        ( 
          { 'label'  : 'Configure', 
            'action' : 'manage_configure',
            'help'   : ('STLDAPManager', 'Add.stx' ) },
          { 'label'  : 'People',
            'action' : 'manage_people',
            'help'   : ('STLDAPManager', 'Object.stx' ) },
          { 'label'  : 'Org Units',
            'action' : 'manage_orgUnits',
            'help'   : ('STLDAPManager', 'Object.stx' ) },
          { 'label'  : 'Groups',
            'action' : 'manage_groups',
            'help'   : ('STLDAPManager', 'Object.stx' ) },
          { 'label'  : 'Schema',
            'action' : 'manage_schema',
            'help'   : ('STLDAPManager', 'Schema.stx' ) },
          { 'label'  : 'Searching & OrgChart',
            'action' : 'manage_searching',
            'help'   : ('STLDAPManager', 'Searching.stx' ) },
          { 'label'  : 'Save Settings',
            'action' : 'manage_saveSettings',
            'help'   : ('STLDAPManager', 'SaveSettings.stx' ) }, )
        )

    manageGroupPermission =   'LDAP Manage Groups'
    manageOrgUnitPermission = 'LDAP Manage Org Units'
    managePeoplePermission =  'LDAP Manage People'
    manageSchemaPermission =  'LDAP Manage Schema'

    security = ClassSecurityInfo()

    security.declareProtected( 'View', 'index_html')
    security.declareProtected( manageGroupPermission, 'manage_groups' )
    security.declareProtected( manageOrgUnitPermission, 'manage_orgUnits')
    security.declareProtected( managePeoplePermission, 'manage_people' )
    security.declareProtected( manageSchemaPermission, 'manage_saveSettings')
    security.declareProtected( manageSchemaPermission, 'manage_schema')
    security.declareProtected( manageSchemaPermission, 'manage_searching')
    security.declareProtected( manageSchemaPermission, 'manage_configureForm')
    security.declareProtected( manageSchemaPermission, 'manage_configure')

    manage_configureForm = DTMLFile('dtml/configure', globals())
    manage_main = manage_configureForm
    orgChartForm = DTMLFile('dtml/orgChart', globals())
    searchForm = DTMLFile('dtml/search', globals())
    viewForm = DTMLFile('dtml/view', globals())

    def __init__( self, id, title, ldapServer, directoryBase, personSearchBase, 
                  orgUnitBrowseBase, groupBrowseBase, zopeRoleSearchBase, 
                  zopeRoleSearchScope, personNamingAttribute, 
                  personsOUAttr, givenAttr, groupObjectClass,
                  REQUEST=None ):

        self.id = id
        self.title = title
        self.LDAPDirectory = \
            LDAPDirectory( ldapServer, directoryBase, personSearchBase, 
                           orgUnitBrowseBase, groupBrowseBase, 
                           zopeRoleSearchBase, zopeRoleSearchScope,
                           personNamingAttribute, personsOUAttr,
                           givenAttr, groupObjectClass )
        user = self._getUser( self.LDAPDirectory, REQUEST )
        self.schema = self.LDAPDirectory._personManagedSchema

        #   For each object category, create a temporary ZSchema object, 
        #   and use it to update the schema with information from LDAP,
        #   and to sync the list of object controllers with the permission
        #   associated with that category.

        zSchema = ZSchema( user, {'objectCategory':'person'} )
        zSchema = zSchema.__of__( self )
        zSchema.updateSchemaFromLDAP(user)
        zSchema._setPermissionName( self.managePeoplePermission )
        zSchema._assignRolesToPermission()

        zSchema = ZSchema( user, {'objectCategory':'orgUnit'} )
        zSchema = zSchema.__of__( self )
        zSchema.updateSchemaFromLDAP(user)
        zSchema._setPermissionName( self.manageOrgUnitPermission )
        zSchema._assignRolesToPermission()

        zSchema = ZSchema( user, {'objectCategory':'group'} )
        zSchema = zSchema.__of__( self )
        zSchema.updateSchemaFromLDAP(user)
        zSchema._setPermissionName( self.manageGroupPermission )
        zSchema._assignRolesToPermission()

    def index_html( self, REQUEST, message=None ):
        """
        Display the appropriate object management form for 
        the 'person' objectCategory, after handling any update
        information that may have been in the request.
        """
        user = self._getUser( self.LDAPDirectory, REQUEST )
        if 'AUTH' in user.getRoles():
            REQUEST.set( 'objectCategory', 'person' )
            zObject = ZObject( user, REQUEST, message )
            zObject = zObject.__of__( self )
            return zObject.showObjectForm()
        else:
            REQUEST.set( 'user', AnonymousUser(self.LDAPDirectory) )
            if not REQUEST.has_key('formID'):
                REQUEST.set( 'formID', 'search' )
            return self.viewForm( self, REQUEST )

    def login(self, REQUEST, RESPONSE):
        """ Allow user to authenticate. """
        user = self._getUser( self.LDAPDirectory, REQUEST )
        if 'AUTH' in user.getRoles():
            url = REQUEST.get( 'URL1' ) + '/'
            RESPONSE.redirect( url )
        else:
            realm = RESPONSE.realm
            RESPONSE.setStatus(401)
            RESPONSE.setHeader('WWW-Authenticate',
                               'basic realm=%s' % realm, 1 )

    def manage_configure( self, REQUEST=None):
        """
        Manage STLDAPManager properties.

        """
        message = ''
        if REQUEST.has_key('updateConfiguration'):
            if not REQUEST.has_key('cancel'):
                self.title = REQUEST.get( 'title', '' )

                ldapServer = REQUEST.get( 'ldapServer' )
                directoryBase = REQUEST.get( 'directoryBase' )
                personSearchBase = REQUEST.get( 'personSearchBase' )
                orgUnitBrowseBase = REQUEST.get( 'orgUnitBrowseBase' )
                groupBrowseBase = REQUEST.get( 'groupBrowseBase' )
                zopeRoleSearchBase = REQUEST.get( 'zopeRoleSearchBase' )
                zopeRoleSearchScope = REQUEST.get( 'zopeRoleSearchScope' )
                personNamingAttribute = REQUEST.get( 'personNamingAttribute' )
                personsOUAttr = REQUEST.get( 'personsOUAttr' )
                givenAttr = REQUEST.get( 'givenAttr' )
                groupObjectClass = REQUEST.get( 'groupObjectClass' )

                self.LDAPDirectory.setServer( ldapServer )
                self.LDAPDirectory.setDirectoryBase( directoryBase )

                if personSearchBase:
                    self.LDAPDirectory.setPersonSearchBase( personSearchBase )
                else:
                    self.LDAPDirectory.setPersonSearchBase( ldapDirectoryBase )

                if orgUnitBrowseBase:
                    self.LDAPDirectory.setOrgUnitBrowseBase( orgUnitBrowseBase )
                else:
                    self.LDAPDirectory.setOrgUnitBrowseBase( ldapDirectoryBase )

                if groupBrowseBase:
                    self.LDAPDirectory.setGroupBrowseBase( groupBrowseBase )
                else:
                    self.LDAPDirectory.setGroupBrowseBase( ldapDirectoryBase )

                if zopeRoleSearchBase:
                    self.LDAPDirectory.setZopeRoleSearchBase(zopeRoleSearchBase)
                else:
                    self.LDAPDirectory.setZopeRoleSearchBase( directoryBase )

                self.LDAPDirectory.setZopeRoleSearchScope(zopeRoleSearchScope)

                personSchema = self.LDAPDirectory._personManagedSchema
                if personNamingAttribute != '':
                    personSchema.setNamingAttribute(personNamingAttribute)
                if personsOUAttr == 'dn':
                    personSchema.setPersonsOUAttr(personsOUAttr)
                elif personsOUAttr == 'useGivenAttr':
                    personSchema.setPersonsOUAttr(givenAttr)

                groupSchema = self.LDAPDirectory._groupManagedSchema
                objectClassAttr = groupSchema.getAttributeSchema('objectClass')
                if groupObjectClass in ['groupOfNames','groupOfUniqueNames']:
                    objectClasses = ['top',groupObjectClass]
                    objectClassAttr.setRequiredValues( objectClasses )

                message = 'Updated configuration.'

        return self.manage_configureForm( self, REQUEST,
                                          manage_tabs_message=message)

    def manage_groups( self, REQUEST, message=None ):
        """
        Display the appropriate object management form for 
        the 'group' objectCategory, after handling any update
        information that may have been in the request.
        """
        user = self._getUser( self.LDAPDirectory, REQUEST )
        REQUEST.set( 'objectCategory', 'group' )
        zObject = ZObject( user, REQUEST, message )
        zObject = zObject.__of__( self )
        return zObject.showObjectForm()

    def manage_orgUnits( self, REQUEST, message=None ):
        """
        Display the appropriate object management form for 
        the 'orgUnit' objectCategory, after handling any update
        information that may have been in the request.
        """
        user = self._getUser( self.LDAPDirectory, REQUEST )
        REQUEST.set( 'objectCategory', 'orgUnit' )
        zObject = ZObject( user, REQUEST, message )
        zObject = zObject.__of__( self )
        return zObject.showObjectForm()

    def manage_people( self, REQUEST, message=None ):
        """
        Display the appropriate object management form for 
        the 'person' objectCategory, after handling any update
        information that may have been in the request.
        """
        user = self._getUser( self.LDAPDirectory, REQUEST )
        REQUEST.set( 'objectCategory', 'person' )
        zObject = ZObject( user, REQUEST, message )
        zObject = zObject.__of__( self )
        return zObject.showObjectForm()

    def manage_saveSettings( self, REQUEST, message=None ):
        """
        Display a form that allows the user to save configuration
        settings in a number of formats:  1) security settings in
        OpenLDAP slapd.conf format, 2) security settings as iPlanet
        ACI's in LDIF format, or 3) the entire configuration in
        SchemaDefaults format.
        """
        REQUEST.set('schemaFormID','saveSettings')
        user = self._getUser( self.LDAPDirectory, REQUEST )
        zSchema = ZSchema( user, REQUEST )
        zSchema = zSchema.__of__( self )
        return zSchema.showSchemaForm( message )

    def manage_schema( self, REQUEST, message=None ):
        """
        Display the appropriate schema management form, after 
        handling any update information that may have been in 
        the request.
        """
        user = self._getUser( self.LDAPDirectory, REQUEST )
        zSchema = ZSchema( user, REQUEST )
        zSchema = zSchema.__of__( self )
        return zSchema.showSchemaForm( message )

    def manage_searching( self, REQUEST, message=None ):
        """
        Display the form for managing the public searching/org
        chart view of this object.
        """
        REQUEST.set('schemaFormID','schemaSearching')
        user = self._getUser( self.LDAPDirectory, REQUEST )
        zSchema = ZSchema( user, REQUEST )
        zSchema = zSchema.__of__( self )
        return zSchema.showSchemaForm( message )

    def orgChart( self, REQUEST, message=None ):
        """
        Given a REQUEST containing a distinguished name, display it.
        Otherwise show the search form.

        """
        #user = AnonymousUser( self.LDAPDirectory )
        user = self._getUser( self.LDAPDirectory, REQUEST )
        REQUEST.set('user',user)
        REQUEST.set( 'formID', 'orgChart' )
        return self.orgChartForm( self, REQUEST )

    def search( self, REQUEST, message=None ):
        """
        Show the search form.

        """
        #user = AnonymousUser( self.LDAPDirectory )
        user = self._getUser( self.LDAPDirectory, REQUEST )
        REQUEST.set('user',user)
        REQUEST.set( 'formID', 'search' )
        return self.searchForm( self, REQUEST )

    def _getUser(self, ldapDirectory, REQUEST):
        """
            Return an AuthenticatedUser connection object in the case of
            an LDAP authenticated user.  A Zope user object created by
            one of Jens' products will have a 'distinguishedName' 
            attribute.  Otherwise, return an AnonymousUser connection.

            But because authenticated users can 
            change their own passwords, the Zope user object can be holding 
            on to invalid credentials.  If that is the case, then expire 
            the Zope user, and return an AnonymousUser connection.  Users
            not already logged in also get an AnonymousUser connection.
        """

        user = getSecurityManager().getUser()

        if hasattr( user, 'getUserDN' ):
            dn = user.getUserDN()
            pw = user._getPassword() 
            try:
                LDAPUser = AuthenticatedUser( dn, pw, ldapDirectory,
                                              zopeUser=user)
            except INVALID_CREDENTIALS:
                user._expire()
                raise 'Unauthorized'
            except:
                LDAPUser = AnonymousUser( ldapDirectory, zopeUser=user )
        else:            
            LDAPUser = AnonymousUser( ldapDirectory, zopeUser=user )
        return LDAPUser

def addForm( self, client=None, REQUEST={}, RESPONSE=None, 
             manage_tabs_message=None, **kw ):
    """
    Show the STLDAPManager add form, taking defaults from SchemaDefaults.

    """
    import SchemaDefaults
    self.addFormSource = DTMLFile("dtml/addForm", globals(),
                                  form_title='Add ST LDAP Manager',
                                  help_product='STLDAPManager',
                                  help_topic='Add.stx')

    REQUEST.set( 'ldapServer', SchemaDefaults.defaultLDAPServer )
    REQUEST.set( 'directoryBase', SchemaDefaults.defaultDirectoryBase )
    REQUEST.set( 'personSearchBase', SchemaDefaults.defaultPersonSearchBase )
    REQUEST.set( 'orgUnitBrowseBase', SchemaDefaults.defaultOrgUnitBrowseBase )
    REQUEST.set( 'groupBrowseBase', SchemaDefaults.defaultGroupBrowseBase )
    REQUEST.set( 'zopeRoleSearchBase', 
                     SchemaDefaults.defaultZopeRoleSearchBase )
    REQUEST.set( 'zopeRoleSearchScope',
                     SchemaDefaults.defaultZopeRoleSearchScope )
    REQUEST.set( 'personNamingAttribute', 
                     SchemaDefaults.defaultUserNamingAttribute )
    REQUEST.set( 'personsOUAttr', SchemaDefaults.defaultPersonsOUAttr )
    REQUEST.set( 'groupObjectClass',
                     SchemaDefaults.defaultGroupObjectClass )
    return self.addFormSource( self, REQUEST,
                               manage_tabs_message=manage_tabs_message )

def addFunction( dispatcher, id, title, ldapServer, directoryBase, 
                 personSearchBase, orgUnitBrowseBase, groupBrowseBase,
                 zopeRoleSearchBase, zopeRoleSearchScope, 
                 personNamingAttribute, personsOUAttr, givenAttr,
                 groupObjectClass, REQUEST=None):
    """ 
    Create a new STLDAPManager and add it to myself.

    """
    # Trap for an empty id or ldapServer.  Redraw the form with an 
    # error message.
    #
    if str(id) == '' and REQUEST is not None:
        message = "You must supply an id for this object."
        return dispatcher.Control_Panel.Products.STLDAPManager.addForm( 
                        dispatcher, REQUEST, manage_tabs_message=message )

    if str(ldapServer) == '' and REQUEST is not None:
        message = """You must supply an LDAP Server.  Zope must be able
                     to connect to the LDAP server to properly initialize
                     this object."""
        return dispatcher.Control_Panel.Products.STLDAPManager.addForm( 
                        dispatcher, REQUEST, manage_tabs_message=message )
    
    inst = STLDAPManager( id, title, ldapServer, directoryBase, 
                          personSearchBase, orgUnitBrowseBase, 
                          groupBrowseBase, zopeRoleSearchBase, 
                          zopeRoleSearchScope, personNamingAttribute,
                          personsOUAttr, givenAttr, groupObjectClass,
                          REQUEST )
    dispatcher.Destination()._setObject(id, inst)

    if REQUEST is not None:
        dispatcher.manage_main(dispatcher, REQUEST)

InitializeClass(STLDAPManager)
