"""
    ZObject.py
    Author: $Author: payne $
    Date:   $Date: 2002/08/06 15:50:52 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    The ZObject class is a Zope-aware object that instantiates
    an internal LDAP object appropriate for the DN information
    passed in the REQUEST.  It contains methods used for drawing
    the forms to display an LDAP object's attributes, and for
    updating the LDAP object from REQUEST information submitted
    from forms.
"""

__version__ = '$Revision: 1.42 $'[11:-2]

from AccessControl import getSecurityManager
from Acquisition import Implicit
from copy import deepcopy
from DocumentTemplate.DocumentTemplate import HTML
from Globals import DTMLFile, InitializeClass
from string import join, lower
import types
from urllib import quote_plus, unquote_plus

from LDAPExceptions import *
from LDAPObject import LDAPObject
from LDAPUtilities import _ldap, DNNormalize, DNsAreEquivalent, \
    getNameFromDN, listsAreEquivalent, list1ContainsList2, \
    mergeLists, TRUE, FALSE
from SearchLister import SearchLister

class ZObject( Implicit ):

    #   DTML templates for the various "faces" of the Objects tab.
    #   Each template's forms must contain a hidden variable that
    #   identifies the form.  The name is 'objectFormID', and the
    #   value must match one of the tests for it in _manage_objects.
    #
    #     <input type=hidden name=objectFormID value=objectAdd>

    #   DTML templates for the various "faces" of the Search tab.

    manage_objectFormSelection = DTMLFile('dtml/objectFormSelection', globals())
    _browseGroupsForm = DTMLFile('dtml/objectBrowseGroups', globals())
    _browseOrgUnitsForm = DTMLFile('dtml/objectBrowseOrgUnits', globals())
    _browsePeopleForm = DTMLFile('dtml/objectBrowsePeople', globals())
    _displayForm = DTMLFile('dtml/objectDisplay', globals())
    _objectAddForm = DTMLFile('dtml/objectAdd', globals())
    _objectDelForm = DTMLFile('dtml/objectDel', globals())
    _objectGroupMembershipForm = DTMLFile('dtml/objectGroupMembership', globals())
    _objectRemoveMembersForm = DTMLFile('dtml/objectRemoveMembers', globals())
    _personDelForm = DTMLFile('dtml/objectDelPerson', globals())
    _searchForm = DTMLFile('dtml/objectSearch', globals(),
                           management_view='People')
    _selectGroupDNForm = DTMLFile('dtml/selectGroupDN', globals())
    _selectOrgUnitDNForm = DTMLFile('dtml/selectOrgUnitDN', globals())
    _selectPersonDNForm = DTMLFile('dtml/selectPersonDN', globals())
    _selectPersonOrgUnitDNForm = DTMLFile('dtml/selectPersonOrgUnitDN',globals())

    def __init__( self, user, REQUEST, message=None ):

        self.user = user
        self.message = message
        self.request = REQUEST

        #   Get an LDAPObject.  

        self.object = LDAPObject( user )
        distinguishedName = self.request.get( 'distinguishedName', None )

        self._IChangedMyOwnPassword = FALSE

        if distinguishedName is not None:
            distinguishedName = unquote_plus( distinguishedName )
            self.object.refresh( distinguishedName )
            self.objectCategory = self.object.getObjectCategory()
        elif self.request.has_key( 'objectCategory' ):
            self.objectCategory = self.request['objectCategory']
        else:
            self.objectCategory = 'person'

        if lower(self.objectCategory) == 'person':
            self.schema = user.LDAPDirectory.getPersonSchema()
        elif lower(self.objectCategory) == 'orgunit':
            self.schema = user.LDAPDirectory.getOrgUnitSchema()
        elif lower(self.objectCategory) == 'group':
            self.schema = user.LDAPDirectory.getGroupSchema()
        else:
            raise UNRECOGNIZED_OBJECT

        self.namingAttribute = self.schema.getNamingAttribute()
        namingAttributeSchema = \
            self.schema.getAttributeSchema(self.namingAttribute)
        self.friendlyNamingAttribute = namingAttributeSchema.getFriendlyName()

    def showObjectForm(self):

        #   Put into the REQUEST information needed on the forms
       
        message = self.message

        #   If we got here by a form submit button, then we need to
        #   update the directory first.  Any form that needs to hit
        #   the methods in ZObject.updateDirectoryFromRequest must
        #   contain:
        #
        #   <input type="hidden" name="updateObject" value="1">

        if self.request.has_key('updateObject'):
            if not self.request.has_key('cancel'):
                self.message = self.updateDirectoryFromRequest()

                if self.iChangedMyOwnPassword():
                    #   ... then I can't go on.  Expire the Zope user 
                    #   object, and redirect to the search screen,  
                    #    where I can login again.

                    getSecurityManager().getUser()._expire()
                    url = self.request.get('URL1') + '/'
                    self.request.RESPONSE.redirect(url)

            else:
                #   Clean up the REQUEST object by just doing a
                #   redirect back to this form for this object.

                url = "%s?objectFormID=%s" % \
                      ( self.request['URL0'],
                        self.request['objectFormID'] )
                if self.request.has_key('distinguishedName'):
                    url = url + "&distinguishedName=%s" % \
                      ( quote_plus(self.request['distinguishedName']) )
                if self.request.has_key('displayCategory'):
                    url = url + "&displayCategory=%s" % \
                      ( quote_plus(self.request['displayCategory']) )
                self.request.RESPONSE.redirect( url )

        #   Set default form, and render it.  Each form must contain a
        #   hidden tag called 'schemaFormID' whose value, with '_'
        #   prepended, and 'Form' appended, is form rendering DTMLFile
        #   call.
    
        if not self.request.has_key('objectFormID'):
            if self.request.has_key('distinguishedName'):
                self.request.set('objectFormID','display')
            else:
                self.request.set('objectFormID','search')

        if self.request.has_key( 'login' ):
            return self._login()

        elif self.request.has_key( 'logout' ):
            return self._logout()

        elif self.request.objectFormID == 'display':
            return self._displayForm( 
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'addObject':
            return self._objectAddForm( 
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'delObject':
            if self.request['objectCategory'] == 'person':
                return self._personDelForm(
                        self, self.request, manage_tabs_message = self.message )
            else:
                return self._objectDelForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'groupMembership':
            return self._objectGroupMembershipForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'selectPersonOrgUnitDN':
            return self._selectPersonOrgUnitDNForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'selectOrgUnitDN':
            return self._selectOrgUnitDNForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'selectGroupDN':
            return self._selectGroupDNForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'browseOrgUnits':
            return self._browseOrgUnitsForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'browseGroups':
            return self._browseGroupsForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'selectPersonDN':
            return self._selectPersonDNForm( self, self.request )

        elif self.request.objectFormID == 'selectOrgUnitDN':
            return self._selectOrgUnitDNForm( self, self.request )

        elif self.request.objectFormID == 'selectGroupDN':
            return self._selectGroupDNForm( self, self.request )

        elif self.request.objectFormID == 'search':
            return self._searchForm( self, self.request, )

        elif self.request.objectFormID == 'browse':
            return self._browsePeopleForm( self, self.request )

        else:
            self.message = """
                      The previous form's objectFormID does not match a
                      known form display method.
                      """
            return self._objectAddForm(
                        self, self.request, manage_tabs_message = self.message )

    def buildTree(self, startingDN=None,
                  sort="",
                  branchMethod="getChildGroups",
                  action="?",
                  returnsToFormID='display' ):
        """
            Generate a tree view of groups in the directory.
        """
        targetAttribute = self.request.get('targetAttribute','')
        if startingDN is None:
            startingDN = self.LDAPDirectory.getDirectoryBase()

        # The trick is to use the urlparam of the tree tag to
        # hold all of the request variables except for the tree
        # state.  But the tree tag parameters are stupidly parsed
        # as a string, so you can't pass it a DTML variable
        # containing this information.  So I end up generating
        # a string that contains the DTML with the right parameters,
        # create a temporary DocumentTemplate.HTML object, and
        # render it in the context of this method.

        queryString = self.getTreeQuery()

        # Finally, the REQUEST has a key that brings us back
        # to the tree.  We need to remove that key for the version
        # of the query string used in the tree's links.

        queryStringReturn = self.getTreeQueryReturn( returnsToFormID )

        # Here's a string template that holds the DTML returned by
        # this method.  Notice the replacement strings at the end.
        #
        treeString = """
          <dtml-with user>
          <dtml-call "REQUEST.set('mappingDN', '%s')">
          <dtml-tree urlparam="%s"
                     id="mappingDN"
                     sort="%s"
                     single="1"
                     branches_expr="%s(mappingDN)">
            <a href="&dtml-URL0;%s%s&%s=<dtml-var "mappingDN" url_quote>">&dtml-%s;</a>
          </dtml-tree>
          </dtml-with> """ % \
                           ( startingDN,
                             queryString,
                             sort,
                             branchMethod,
                             action,
                             queryStringReturn,
                             targetAttribute,
                             sort
                           )

        #return treeString

        # Make a DocumentTemplate.HTML object out of the string, and
        # render it.
        #
        ob = HTML( treeString )
        return apply(HTML.__call__, ( ob, self, self.request ))

    def canControl( self, objectCategory ):
        """
            Can the user be a controller for objects of the specified
            objectCategory?  That is, does the user have any roles
            in the ManagedSchema's controllers list.
        """
        controllers = self.schema.getControllers()
        #for item in getSecurityManager().getUser().getRoles():
        for item in self.user.getRoles():
            if item in controllers:
                return TRUE
        return FALSE

    def getGroupMembers( self ):
        """
            If the saved LDAPObject has been refreshed with information
            from a groupOfNames or groupOfUniqueNames in the directory,
            then return the list of values (DNs) of the member and/or
            uniqueMember attributes.
        """
        groupMembers = []
        objectClassAttr = self.object.getManagedAttribute('objectClass')
        if 'groupOfNames' in objectClassAttr.getValidValues():
            groupMembers.extend(self.object.getManagedAttribute('member').getValues())
        if 'groupOfUniqueNames' in objectClassAttr.getValidValues():
            groupMembers.extend( self.object.getManagedAttribute('uniqueMember').getValues())
        groupMembers.sort()
        return groupMembers

    def getObject( self ):
        """
            Returns the saved LDAPObject object
        """
        return self.object

    def getObjectClasses( self ):
        """
            Returns the values of the objectClass attribute for the
            saved LDAPObject
        """
        return self._objectClasses

    def getReadableSearchKeys(self):
        """
            Returns a filtered list of search keys, containing only
            those readable by the user.
        """
        searchKeys = deepcopy( self.schema._searchKeys )
        for attributeName in searchKeys:
            attributeSchema = self.schema.getAttributeSchema(attributeName)
            readers = attributeSchema.getReaders()
            canRead = FALSE
            for role in self.user.getRoles():
                if role in readers:
                    canRead = TRUE
                    break
            if not canRead:
                searchKeys.remove(attributeName)
        return searchKeys

    def getTreeQuery(self):
        # Build a query string to be used as the urlparam of a
        # tree tag.
        #
        if self.request.form.has_key('queryString'):
            del self.request.form['queryString']
        if self.request.form.has_key('tree-c'):
            del self.request.form['tree-c']
        if self.request.form.has_key('tree-e'):
            del self.request.form['tree-e']

        targetAttribute = self.request.get('targetAttribute','')
        if self.request.form.has_key(targetAttribute):
            del self.request.form[targetAttribute]

        if not self.request.form.has_key( 'selectDN_'+targetAttribute ):
            self.request.form['selectDN_'+targetAttribute] =  ''

        output = self.urlencode( self.request.form, 1 )
        del self.request.form['selectDN_'+targetAttribute]
        return output

    def getTreeQueryReturn( self, returnsToFormID='display' ):
        # Generate a tree query string suitable for returning to the
        # calling form.
        #
        self.getTreeQuery()
        self.request.set('objectFormID', returnsToFormID )
        return self.urlencode( self.request.form, 1 )

    def iChangedMyOwnPassword( self ):
        """
            Returns the value of the attribute that checkes whether 
            I changed my own password.
        """
        return self._IChangedMyOwnPassword

    def selectDNKey(self):
        """
            If the REQUEST contains a key of the form 'selectDN_ldapName'
            this method returns 'ldapName'
        """
        for key in self.request.keys():
            if key[0:9] == 'selectDN_':
                return key[9:]
        return ''

    def updateDirectoryFromRequest(self):
        """
            Make directory updates according to the REQUEST variables.
        """

        #   Invoke a method that matches the name of the submit
        #   button (with '_' prepended).

        if self.request.has_key( 'updateObjectValues' ):
            message = self._updateObjectValues()

        elif self.request.has_key( 'updateMembership' ):
            url = self.request['URL1'] + \
                  "/manage_groups?distinguishedName=" + \
                  "%s" % quote_plus(self.request.distinguishedName) + \
                  "&objectFormID=groupMembership"
            self.request.RESPONSE.redirect( url )
            message = url

        elif self.selectDNKey() != '':
            message = self._selectDN()

        elif self.request.has_key( 'addObject' ):
            message = self._addObject()

        elif self.request.has_key( 'delObject' ):
            message = self._delObject()

        elif self.request.has_key( 'addMembers' ):
            message = self._addMembers()

        elif self.request.has_key( 'removeMembers' ):
            message = self._removeMembers()

        else:
            # You can get here if your form has a single text input
            # field and you press return rather than the form button.
            # Maybe there is a way to figure it out, maybe we don't
            # care.

            message = ""
            if self.request.get('objectFormID','') == 'selectPersonDN':
                message = self._selectDN()

        return message

    # urlencode needs the doseq option that came after Python1.5.2
    # This code comes from Python 2.1.3's urllib.py

    def urlencode(self,query,doseq=0):
        """Encode a sequence of two-element tuples or dictionary into a URL query string.
    
        If any values in the query arg are sequences and doseq is true, each
        sequence element is converted to a separate parameter.
    
        If the query arg is a sequence of two-element tuples, the order of the
        parameters in the output will match the order of parameters in the
        input.
        """
    
        if hasattr(query,"items"):
            # mapping objects
            query = query.items()
        else:
            # it's a bother at times that strings and string-like objects are
            # sequences...
            try:
                # non-sequence items should not work with len()
                x = len(query)
                # non-empty strings will fail this
                if len(query) and type(query[0]) != types.TupleType:
                    raise TypeError
                # zero-length sequences of all types will get here and succeed,
                # but that's a minor nit - since the original implementation
                # allowed empty dicts that type of behavior probably should be
                # preserved for consistency
            except TypeError:
                ty,va,tb = sys.exc_info()
                raise TypeError, "not a valid non-string sequence or mapping object", tb
    
        l = []
        if not doseq:
            # preserve old behavior
            for k, v in query:
                k = quote_plus(str(k))
                v = quote_plus(str(v))
                l.append(k + '=' + v)
        else:
            for k, v in query:
                k = quote_plus(str(k))
                if type(v) == types.StringType:
                    v = quote_plus(v)
                    l.append(k + '=' + v)
#                elif type(v) == types.UnicodeType:
#                    # is there a reasonable way to convert to ASCII?
#                    # encode generates a string, but "replace" or "ignore"
#                    # lose information and "strict" can raise UnicodeError
#                    v = quote_plus(v.encode("ASCII","replace"))
#                    l.append(k + '=' + v)
                else:
                    try:
                        # is this a sufficient test for sequence-ness?
                        x = len(v)
                    except TypeError:
                        # not a sequence
                        v = quote_plus(str(v))
                        l.append(k + '=' + v)
                    else:
                        # loop over the sequence
                        for elt in v:
                            l.append(k + '=' + quote_plus(str(elt)))
        return join(l,'&')
    
    
    # Internal methods

    def _addObject(self):

        #   Combine items from the REQUEST to form the distinguished
        #   name and object category of the object to be added.  Add
        #   and redirect to a display of it.

        namingAttributeValue = self.request.get( 'namingAttributeValue' )
        parentDN = unquote_plus( self.request.get( 'parentDN' ) )
        objectCategory = self.request.get( 'objectCategory' )

        distinguishedName = "%s=%s,%s" % ( self.namingAttribute,
                                           namingAttributeValue,
                                           parentDN )

        try:
            self.object.create( distinguishedName, objectCategory )
            message = "Added object '%s'.  " % ( distinguishedName )
            url = "%s?distinguishedName=%s&manage_tabs_message=%s" % \
                  ( self.request['URL0'], 
                    quote_plus(distinguishedName), 
                    quote_plus(message) )
            self.request.RESPONSE.redirect( url )

        except ALREADY_EXISTS:
            message = """Object '<a href="?distinguishedName=%s">%s</a>' 
                already exists!""" % ( distinguishedName, distinguishedName )
            return message

        # any other exception will raise a Zope error

    def _delObject(self):
        if not self.request.has_key('areYouSure'):
            self.request.set( 'areYouSure', '1' )
            return "ARE YOU SURE?"

        elif self.request.has_key('ids'):
            ids = self.request.get( 'ids', [] )
            for distinguishedName in ids:
                self.object.refresh( distinguishedName )
                self.object.delete( self.request.get('areYouSure') )
        else:
            targetDN = self.request.get('targetDN') 
            self.object.refresh( targetDN )
            self.object.delete( self.request.get('areYouSure') )

        message = "Objects deleted."

        url = "%s?objectFormID=delObject" % ( self.request['URL0'] )
        self.request.RESPONSE.redirect( url )

    def _addMembers(self):
        ids = self.request.get( 'ids', [] )
        # Am I to add a 'member' or 'uniqueMember'?

        objectClassAttr = self.object.getManagedAttribute('objectClass')
        if 'groupOfNames' in objectClassAttr.getRequiredValues():
            memberAttrName = 'member'
        elif 'groupOfUniqueNames' in objectClassAttr.getRequiredValues():
            memberAttrName = 'uniqueMember'
        else:
            memberAttrName = None
        if memberAttrName:
            memberAttr = self.object.getManagedAttribute(memberAttrName)

        for distinguishedName in ids:
            memberAttr.addValue( distinguishedName )

        if len( ids ) > 0:
            self.object.updateDirectory()
            message = "Added group members %s" % ids
        else:
            message = ""

        self.request.set( 'ids', [] )
        return message

    def _login(self):
        if 'AUTH' in self.user.getRoles():
            self.request.RESPONSE.redirect( self.request.get( 'URL0' ) )
        else:
            realm = self.request.RESPONSE.realm
            self.request.RESPONSE.setStatus(401)
            self.request.RESPONSE.setHeader('WWW-Authenticate',
                                            'basic realm=%s' % realm )

    def _logout(self):
        realm = self.request.RESPONSE.realm
        self.request.RESPONSE.setStatus(401)
        self.request.RESPONSE.setHeader('WWW-Authenticate',
                                        'basic realm=%s' % realm, 1 )
        self.request.RESPONSE.setBody("""<html>
<head><title>Logout</title></head>
<body>
<p>
You have been logged out.
</p>
</body>
</html>""")
        return

    def _removeMembers(self):
        if not self.request.get('areYouSure',0):
            self.request.set( 'areYouSure', 1 )
            return "ARE YOU SURE?"

        elif self.request.has_key('ids'):
            self.request.set('areYouSure',0)
            objectClassAttr = self.object.getManagedAttribute('objectClass')
            if 'groupOfNames' in objectClassAttr.getRequiredValues():
                memberAttr = self.object.getManagedAttribute('member')
            elif 'groupOfUniqueNames' in objectClassAttr.getRequiredValues():
                memberAttr = self.object.getManagedAttribute('uniqueMember')
            else:
                memberAttr = None
            ids = self.request.get( 'ids', [] )
            if memberAttr is not None:
                for distinguishedName in ids:
                    memberAttr.delValue( distinguishedName )

            if len( ids ) > 0:
                self.object.updateDirectory()
                message = "Removed group members"
            else:
                message = ""

        self.request.set( 'ids', [] )
        return message

    def _selectDN(self):

        targetAttribute = self.request.get('targetAttribute',self.selectDNKey())
        option = self.request.get( 'selectDN_' + targetAttribute )
        objectCategory = self.request.get( 'objectCategory' )

        #   If we got here from the _selectDNForm then update the REQUEST.
        #   Furthermore, refresh my object with the selected DN, if
        #   we were selecting an OrgUnit or group, which are selected
        #   singly.

        if option in [" Select "," Nobody "]:
            if option == " Select ":
                self.request.set( targetAttribute, self.request.get( 'ids' ) )
            elif option == " Nobody ":
                self.request.set( targetAttribute, [] )

            if self.request.form.has_key('targetAttribute'):
                del self.request.form['targetAttribute']
            self.request.set( 'objectFormID', 'display' )
            return "Selected %s.  You still need to Save." % ( targetAttribute )

        else:
            self.request.set( 'targetAttribute', targetAttribute )

            if self.request.has_key( 'pointsTo' ):
                pointsTo = self.request['pointsTo']

            elif hasattr( self, 'schema' ) and \
             targetAttribute in self.schema.getManagedAttributeNames():
                attrSchema = self.schema.getAttributeSchema(targetAttribute)
                pointsTo = attrSchema.getPointsTo()

            else:
                pointsTo = 'orgUnit'

            #   Set REQUEST variables so that one of the
            #   _select(objectCategory)DNForm gets
            #   displayed next.

            if lower(pointsTo) == 'person':
                self.request.set('objectFormID', 'selectPersonDN' )
            elif lower(pointsTo) == 'personorgunit':
                self.request.set('objectFormID', 'selectPersonOrgUnitDN' )
            elif lower(pointsTo) == 'orgunit':
                self.request.set('objectFormID', 'selectOrgUnitDN' )
            elif lower(pointsTo) == 'group':
                self.request.set('objectFormID', 'selectGroupDN' )
            return self.request['objectFormID']

    def _updateObjectValues(self):

        #   Pull out all of the ManagedAttributes in this displayCategory
        #   that the user can write.  Update them with values from the
        #   REQUEST, but them back, and update the directory.

        distinguishedName = self.request.get( 'distinguishedName' )
        displayCategory = self.request.get( 'displayCategory' )
        object = self.object
        object.refresh()

        attrNames = object.getManagedAttributeNames( displayCategory )
        attrNames = filter( lambda i,o=object: o.canWriteAttribute(i),
                            attrNames )

        for attrName in attrNames:
            attrObject = object.getManagedAttribute( attrName )

            #   Special case for paswords, which need two values in
            #   the REQUEST.

            if attrObject.getSyntaxName() == 'Password':
                pw1 = self.request.get( attrName+'1', '' )
                pw2 = self.request.get( attrName+'2', '' )

                if pw1 != pw2:
                    self.request.set( attrName+'1', 'Password' )
                    self.request.set( attrName+'2', 'Password' )
                    return "Passwords do not match.  Try again."

                elif pw1 == 'Password':
                    break

                else:
                    attrObject.setValues( [pw1,] )

                    # Did I change my own password?

                    myDN = getSecurityManager().getUser().getUserDN()
                    if DNsAreEquivalent( myDN, distinguishedName ):
                        self._IChangedMyOwnPassword = TRUE

            else:
                attrObject.setValues( self.request.get( attrName, [] ) )

            object.setManagedAttribute( attrObject )

        object.updateDirectory()
        return "ZObject: Updated values in the directory."
