"""

    Module: $RCSfile: ZSchema.py,v $
    Author: $Author: payne $
    Date:   $Date: 2002/08/06 15:50:52 $

    The ZSchema class is the mutable schema object class.  It holds
    an AuthenticatedUser instance, a ManagedSchema class instance,
    a ManagedRoles instance, the REQUEST, and contains methods used 
    on the various faces of the Schema tab.

    Schema management is split into three parts:  the schemas for
    'person', 'group', and 'orgUnit' types of object are managed
    separately.  The Schema tab provides a way to select which
    type of object schema is being managed.  The schema tab actually 
    consists of a number of forms:

      o A form for selecting the objectClass list to be associated
        with the object category.  The "Selected Object Classes" list
        contains the object classes expected of all objects of this
        category in the directory, and is used whenever a new object 
        is created.  The "Available Object Classes" list contains
        object classes values that might be held by objects in the
        directory.

      o A form for selecting the attributes to be managed from the list
        of all attributes available in the available and selected
        object classes, selected on the form just described.

      o A form for managing a list of "Display Categories."  A display
        category is a list of attributes to be managed together on a
        form named for the display category.  By creating display
        categories, and grouping attributes under them, you can
        configure the user interface for managing an object in LDAP.

      o A form for managing lists of roles.  This list of roles is used
        for configuring the individual managed attributes, when
        specifying who has read and write access to the attribute.
        There is a list of special "Permanent" roles, not associated
        with LDAP groups, which can be augmented with "Configured"
        roles, which do correspond with LDAP groups.

      o A form for editing the schema of individual managed attributes,
        allowing the manager to supply properties like a friendly name,
        and a list of valid values, and to specify which Zope roles
        have read and write access to the attribute.

      o A form for specifying how to search for objects of this category,
        and how to display the results.  This is done by specifying an 
        LDAP search filter for selecting objects to be queried, and then 
        a list of attributes, whose values are searched for a match to
        a search string.  This list of attributes is also used for
        displaying the results as a table, with control over the column
        headings and the column to be used for sorting.

      o A form for specifying which roles can create and delete objects
        in LDAP.

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

"""
__version__ = '$Revision: 1.55 $'[11:-2]

from AccessControl import ClassSecurityInfo
from AccessControl.Role import RoleManager
from Acquisition import Implicit
from Globals import DTMLFile
from string import join, lower

from LDAPExceptions import REQUIRED_ATTRIBUTE, NO_SUCH_ATTRIBUTE
from LDAPUtilities import _ldap, DNsAreEquivalent, \
    list1ContainsList2, mergeLists, TRUE, FALSE
from LDIFObjectParser import LDIFObjectParser
from LDIFAttributeParser import LDIFAttributeParser
from ManagedAttributeSchema import ManagedAttributeSchema
from ZRoles import ZRoles

class ZSchema( Implicit, RoleManager ):

    security = ClassSecurityInfo()
    security.declareProtected('LDAP Manage Schema', 
        'manage_schemaFormSelection',
        )

    manage_schemaFormSelection = DTMLFile('dtml/schemaFormSelection', globals())

    #   DTML templates for the various "faces" of the Schema tab.
    #   Each template's forms must contain a hidden variable that
    #   identifies the form.  The name is 'schemaFormID', and the
    #   value must match one of the following after prepending an
    #   underscore, and appending 'Form', as in:
    #
    #     <input type=hidden name=schemaFormID value=schemaAttributes>
 
    _schemaAttributesForm = DTMLFile('dtml/schemaAttributes', globals())
    _schemaAttrsByCategoryForm = DTMLFile('dtml/schemaAttrsByCategory',
                                          globals())
    _schemaAttrsByRoleForm = DTMLFile('dtml/schemaAttrsByRole', globals())
    _schemaCategoriesForm = DTMLFile('dtml/schemaCategories', globals())
    _schemaControllersForm = DTMLFile('dtml/schemaControllers', globals())
    _schemaManagedForm = DTMLFile('dtml/schemaManaged', globals())
    _schemaObjectClassesForm = DTMLFile('dtml/schemaObjectClasses',
                                        globals())
    _schemaRenameCategoriesForm = DTMLFile('dtml/schemaRenameCategories', 
                                           globals())
    _schemaRenameRolesForm = DTMLFile('dtml/schemaRenameRoles', globals())
    _schemaRolesForm = DTMLFile('dtml/schemaRoles', globals())
    _schemaSearchingForm = DTMLFile('dtml/schemaSearching',globals(),
                                    management_view='Searching & OrgChart')
    _saveSettingsForm = DTMLFile('dtml/saveSettings',globals())

    def __init__(self, user, REQUEST ):

        #   Update the Zope schema object with values from the LDAP 
        #   schema (and save the list of available LDAP object classes).  
        
        self.user = user
        self.request = REQUEST

        if not self.request.has_key('objectCategory'):
            self.request.set('objectCategory','person')
        category = lower(self.request['objectCategory'])

        if category == 'group':
            self._schema = user.getLDAPDirectory().getGroupSchema()
        elif category == 'person':
            self._schema = user.getLDAPDirectory().getPersonSchema()
        elif category == 'orgunit':
            self._schema = user.getLDAPDirectory().getOrgUnitSchema()
        else:
            assert FALSE            

        self._ldapObjects = self.updateSchemaFromLDAP( user )
        self._managedRoles = user.getLDAPDirectory().getManagedRoles()

    def checkObjectClassCompleteness( self, objectClasses ):
        """
            Given a list of objectClasses, check that the list is
            complete, in the sense that all of the superclasses
            for each item in the list are also in the list.
        """
        for className in objectClasses:
            superClasses = getLDAPSuperClassNames( className,
                                                   self._ldapObjects )
            for item in superClasses:
                if item not in objectClasses:
                    return FALSE
        return TRUE

    def DNsAreEquivalent( self, dn1, dn2):
        """
           Wrapper for LDAPUtilities.DNsAreEquivalent
        """
        return DNsAreEquivalent( dn1, dn2 )

    def getManagedRoles(self):
        """
            Returns a managed roles object.
        """
        return self._managedRoles

    def getManagedSchema(self):
        """
            Returns a managed schema object,
        """
        return self._schema

    def getSecurityAttrConf( self, roles, schema, dn, filter ):
        """
        Given a schema and a list roles, create OpenLDAP slapd.conf
        lines describing the access controls.
        """
        # Build a dictionary keyed by access rules, to sort all of the
        # attributes with the same access rules into the same security
        # assertion.

        dict = self.getSecurityAttrDict( roles, schema )

        # For each key, build a complete OpenLDAP security assertion,
        # converting the key into assertion text.

        zopeRoleSearchBase = \
            self.user.getLDAPDirectory().getZopeRoleSearchBase()
        output = ''

        for key in dict.keys():
            attrNameList = dict[key]
            writeKey = key>>len(roles)
            readKey = key - (writeKey<<len(roles))
            output = output + """
access to dn.subtree="%s"
          filter=%s
          attrs=%s""" % ( dn, filter, join(attrNameList,',') )

            # Do the writers, then the readers.  If neither, then
            # grant 'auth' permission.

            for role in roles:
                mask = 1<<(len(roles) - roles.index(role) - 1 )
                writeFlag = mask & writeKey
                if writeFlag:
                    if role == 'SELF':
                        output = output + """
          by self write """
                    elif role == 'MNGR':
                        pass
#                        output = output + """
#          by dnattr=manager writes"""
                    elif role == 'IFALLOWED':
                        pass
                    elif role not in ['AUTH','ANON']:
                        for group in self.user.getMatchingGroups( 
                          role, zopeRoleSearchBase ):
                            if self.user.whatKindOfGroup(group) == \
                              'groupOfUniqueNames':
                                output = output + """
          by group/groupOfUniqueNames/uniqueMember="%s" write""" % group
                            else:
                                output = output + """
          by group="%s" write""" % group
                    elif role == 'AUTH':
                        output = output + """
          by users write"""
                    elif role == 'ANON':
                        output = output + """
          by anonymous write""" 

            for role in roles:
                mask = 1<<(len(roles) - roles.index(role) - 1 )
                writeFlag = mask & writeKey
                readFlag = mask & readKey
                if readFlag and not writeFlag:
                    if role == 'SELF':
                        output = output + """
          by self read"""
                    elif role == 'MNGR':
                        output = output + """
          by dnattr=manager read"""
                    elif role == 'IFALLOWED':
                        pass
                    elif role not in ['AUTH','ANON']:
                        for group in self.user.getMatchingGroups( 
                          role, zopeRoleSearchBase ):
                            if self.user.whatKindOfGroup(group) == \
                              'groupOfUniqueNames':
                                output = output + """
          by group/groupOfUniqueNames/uniqueMember="%s" read""" % group
                            else:
                                output = output + """
          by group="%s" read""" % group
                    elif role == 'AUTH':
                        output = output + """
          by users read"""
                    elif role == 'ANON':
                        output = output + """
          by anonymous read"""

                elif not writeFlag:
                    if role == 'SELF':
                        output = output + """
          by self auth""" 
                    elif role == 'AUTH':
                        output = output + """
          by users auth"""
                    elif role == 'ANON':
                        output = output + """
          by anonymous auth"""
            output = output+'\n'

        return output

    def getSecurityAttrDict( self, roles, schema ):
        """
        Build a dictionary whose key encodes read/write access by all
        of the permanent roles and the roles that map to LDAP groups,
        and whose value is a list of attribute ldapNames that have the
        encoded permissions.
        """
        dict = {}
        for attrName in schema.getManagedAttributeNames():
            attr = schema.getAttributeSchema( attrName )
            writers = attr.getWriters()
            readers = attr.getReaders()
            writeKey = 0
            readKey = 0
            for role in roles:
                if role in writers:
                    writeKey = (writeKey<<1) + 1
                    readKey = (readKey<<1)
                elif role in readers:
                    writeKey = (writeKey<<1)
                    readKey = (readKey<<1) + 1
                else:
                    writeKey = (writeKey<<1)
                    readKey = (readKey<<1)

            key = (writeKey<<len(roles)) + readKey

            if dict.has_key( key ):
                dict[key].append( attrName )
            else:
                dict[key] = [attrName]
        return dict

    def getSecurityControllerConf(self, schema, dn, filter ):
        """
        Write OpenLDAP slapd.conf lines describing what the schema
        controllers can do.
        """
        output = """
access to dn.subtree="%s"
          filter=%s
          attrs=children,entry""" % ( dn, filter )

        for controller in schema.getControllers():
            for controllerDN in self.user.getMatchingGroups( controller, dn ):
                if self.user.whatKindOfGroup(controllerDN) == 'groupOfUniqueNames':
                    output = output + """
          by group/groupOfUniqueNames/uniqueMember="%s" write""" % controllerDN
                else:
                    output = output + """
          by group="%s" write""" % controllerDN

        output = output + """
          by * read\n"""
        return output

    def getSecurityOpenLDAPConf(self):
        """
        Return the configured security (access) settings in a format
        suitable for use in OpenLDAP's slapd.conf file.
        """
        # Make sure everyone can read the schema

        output = """# Start of LDAP Manager generated configuration

access to dn="cn=subschema" by * read\n"""

        # We need to arrange the available roles in order of most
        # specific to least specific.  We cheat here, and put the
        # permanent roles in order, assuming that we know what they
        # are, rather than taking them from the configuration -- BAD.
        # We put the configured roles just ahead of 'AUTH', and
        # ignore 'IFALLOWED'.

        roles = ['SELF','MNGR',] + \
                self._managedRoles.getConfiguredRoleNames() + \
                ['AUTH','ANON']

        # By object category, we make use of a dictionary whose key 
        # encodes read/write access by all of the permanent roles and 
        # the roles that map to LDAP groups, and whose value is a list
        # of attribute ldapNames that have the encoded permissions.

        # 'person'
        schema = self.user.LDAPDirectory._personManagedSchema
        dn = self.user.LDAPDirectory.getPersonSearchBase()
        filter = '(objectclass=person)'
        output = output + self.getSecurityControllerConf( schema, dn, filter )
        output = output + self.getSecurityAttrConf( roles, schema, dn, filter )

        # 'orgUnit'
        schema = self.user.LDAPDirectory._orgUnitManagedSchema
        dn = self.user.LDAPDirectory.getOrgUnitBrowseBase()
        filter = '(objectclass=organizationalunit)'
        output = output + self.getSecurityControllerConf( schema, dn, filter )
        output = output + self.getSecurityAttrConf( roles, schema, dn, filter )

        # 'group'
        schema = self.user.LDAPDirectory._groupManagedSchema
        dn = self.user.LDAPDirectory.getGroupBrowseBase()
        filter = '(|(objectclass=groupofnames)(objectclass=groupofuniquenames))'
        output = output + self.getSecurityControllerConf( schema, dn, filter )
        output = output + self.getSecurityAttrConf( roles, schema, dn, filter )

        output = output + """
# End of LDAP Manager generated configuration"""
        return output

    def getSecurityIPlanetLDIF(self):
        """
            Returns the configured security (access) settings in LDIF
            format suitable for use iPlanet Directory Server aci's.
        """

        output = ''

        # First, the readable/writeable attributes for the permanent 
        # roles and the roles that map to LDAP groups, and then the
        # controllers, by objectCategory...  Note that the 'IFALLOWED' 
        # role is not implemented.
        # 'person'

        schema = self.user.LDAPDirectory._personManagedSchema
        dn = self.user.LDAPDirectory.getPersonSearchBase()
        filter = '(objectclass=person)'
        objectCategory = 'person'
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'SELF', 'userdn', 'ldap:///self' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'ANON', 'userdn', 'ldap:///anyone' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'AUTH', 'userdn', 'ldap:///all' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'MNGR', 'userattr', 'manager#USERDN' )
        for role in self._managedRoles.getConfiguredRoleNames():
            for roleDN in self.user.getMatchingGroups( role, dn ):
                output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                                    objectCategory,
                                                    role, 'groupdn',
                                                    'ldap:///'+roleDN )
        output = self.iPlanetControllerLDIF( output, schema, dn, filter, 
                                             objectCategory )


        # 'group'

        schema = self.user.LDAPDirectory._groupManagedSchema
        dn = self.user.LDAPDirectory.getGroupBrowseBase()
        filter = '(|(objectclass=groupofnames)(objectclass=groupofuniquenames))'
        objectCategory = 'group'
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'SELF', 'userdn', 'ldap:///self' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'ANON', 'userdn', 'ldap:///anyone' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'AUTH', 'userdn', 'ldap:///all' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'MNGR', 'userattr', 'manager#USERDN' )
        for role in self._managedRoles.getConfiguredRoleNames():
            for roleDN in self.user.getMatchingGroups( role, dn ):
                output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                                    objectCategory,
                                                    role, 'groupdn',
                                                    'ldap:///'+roleDN )
        output = self.iPlanetControllerLDIF( output, schema, dn, filter, 
                                             objectCategory )

        # 'orgUnit'

        schema = self.user.LDAPDirectory._orgUnitManagedSchema
        dn = self.user.LDAPDirectory.getOrgUnitBrowseBase()
        filter = '(objectclass=organizationalUnit)'
        objectCategory = 'orgUnit'
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'SELF', 'userdn', 'ldap:///self' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'ANON', 'userdn', 'ldap:///anyone' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'AUTH', 'userdn', 'ldap:///all' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                            objectCategory,
                                            'MNGR', 'userattr', 'manager#USERDN' )
        for role in self._managedRoles.getConfiguredRoleNames():
            for roleDN in self.user.getMatchingGroups( role, dn ):
                output = self.iPlanetAttributeLDIF( output, schema, dn, filter,
                                                    objectCategory,
                                                    role, 'groupdn',
                                                    'ldap:///'+roleDN )
        output = self.iPlanetControllerLDIF( output, schema, dn, filter, 
                                             objectCategory )
        return output

    def getSchemaForSchemaDefaults( self, schema ):
        """
        Write out attribute schemas for the supplied schema in
        SchemaDefaults format.

        """
        output = ""
        for ldapName in schema.getManagedAttributeNames():
            ldapAttr = schema.getAttributeSchema( ldapName )
            output = output + "\n        '%s' : {" % ldapName
            if ldapAttr.getFriendlyName() != ldapName:
                output = output + """
                'friendlyName' : '%s',""" % ldapAttr.getFriendlyName()
            output = output + """
                'readers' : %s,
                'writers' : %s,
                'displayCategories' : %s,""" \
                % ( ldapAttr.getReaders(),
                    ldapAttr.getWriters(),
                    ldapAttr.getDisplayCategories(), )
            if ldapAttr.hasValidValues():
                output = output + """
                'validValues' : %s,""" % ldapAttr.getValidValues()
            if ldapAttr.getRequiredValues() != []:
                output = output + """
                'requiredValues' : %s,""" % ldapAttr.getRequiredValues()
            if ldapAttr.getPointsTo() != '':
                output = output + """
                'pointsTo' : '%s',""" % ldapAttr.getPointsTo()
            if ldapAttr.getRequired():
                output = output + """
                'required' : %s,""" % ldapAttr.getRequired()
            output = output + " },\n"
        output = output + "        }\n"
        return output

    def getSecuritySchemaDefaults(self):
        """
        Write out the security assertions (actually all of the 
        configuration) in a format suitable for replacing the
        SchemaDefaults module.

        """
        directory = self.user.LDAPDirectory
        personSchema = directory.getPersonSchema()
        groupSchema = directory.getGroupSchema()
        orgUnitSchema = directory.getOrgUnitSchema()
        managedRoles = directory.getManagedRoles()

        output = """\"\"\" SchemaDefaults.py, automatically generated. \"\"\"
"""

        output = output + """
defaultLDAPServer = '%s'
defaultDirectoryBase = '%s'
defaultPersonSearchBase = '%s'
defaultOrgUnitBrowseBase = '%s'
defaultGroupBrowseBase = '%s'
defaultZopeRoleSearchBase = '%s'
defaultZopeRoleSearchScope = %s
""" %   ( directory.getServer(),
          directory.getDirectoryBase(),
          directory.getPersonSearchBase(),
          directory.getOrgUnitBrowseBase(),
          directory.getGroupBrowseBase(),
          directory.getZopeRoleSearchBase(),
          directory.getZopeRoleSearchScope(), )

        groupObjectClass = ''
        for objectClass in \
         groupSchema.getAttributeSchema('objectClass').getRequiredValues():
            if objectClass[:5] == 'group':
                groupObjectClass = objectClass
                break

        output = output + """
defaultPersonsOUAttr = '%s'
defaultGroupObjectClass = '%s'
defaultGroupControllers = %s
defaultOrgUnitControllers = %s
defaultUserControllers = %s
""" %   ( personSchema.getPersonsOUAttr(),
          groupObjectClass,
          groupSchema.getControllers(),
          orgUnitSchema.getControllers(),
          personSchema.getControllers(), )
          
        output = output + """
defaultConfiguredRoleNames = %s
defaultCanonicalRoleNames = %s
""" %   ( managedRoles.getConfiguredRoleNames(),
          managedRoles.getPermanentRoles(),)

        output = output + "defaultPermanentRoles = {\n"
        for role in managedRoles.getPermanentRoles():
            output = output + """
        '%s' : {
                'roleName'     : '%s',
                'description'  : \"\"\"%s\"\"\" },
""" %       ( role,
              managedRoles.getRoleNameFromCanonicalName(role),
              managedRoles.getRoleDescriptionFromCanonicalName(role), )
        output = output + "        }\n"

        output = output + """
defaultGroupDisplayCategories = %s
defaultOrgUnitDisplayCategories = %s
defaultUserDisplayCategories = %s
""" %   ( groupSchema.getDisplayCategories(),
          orgUnitSchema.getDisplayCategories(),
          personSchema.getDisplayCategories(), )

        output = output + """
defaultGroupNamingAttribute = '%s'
defaultOrgUnitNamingAttribute = '%s'
defaultUserNamingAttribute = '%s'
""" %   ( groupSchema.getNamingAttribute(),
          orgUnitSchema.getNamingAttribute(),
          personSchema.getNamingAttribute(), )

        output = output + """
defaultGroupSearchFilter = "%s"
defaultGroupSearchKeys = %s
defaultGroupSearchSortKey = '%s'
""" %   ( groupSchema.getSearchFilter(),
          groupSchema.getSearchKeys(),
          groupSchema.getSearchSortKey(), )

        output = output + """
defaultOrgUnitSearchFilter = "%s"
defaultOrgUnitSearchKeys = %s
defaultOrgUnitSearchSortKey = '%s'
""" %   ( orgUnitSchema.getSearchFilter(),
          orgUnitSchema.getSearchKeys(),
          orgUnitSchema.getSearchSortKey(), )

        output = output + """
defaultUserSearchFilter = "%s"
defaultUserSearchKeys = %s
defaultUserSearchSortKey = '%s'
""" %   ( personSchema.getSearchFilter(),
          personSchema.getSearchKeys(),
          personSchema.getSearchSortKey(), )

        output = output + """
defaultPointsToTargets = %s
""" %   ( personSchema.getPointsToTargets(), )

        output = output + "\ndefaultGroupSchema = {\n"
        output = output + self.getSchemaForSchemaDefaults( groupSchema )

        output = output + "\ndefaultOrgUnitSchema = {\n"
        output = output + self.getSchemaForSchemaDefaults( orgUnitSchema )

        output = output + "\ndefaultUserSchema = {\n"
        output = output + self.getSchemaForSchemaDefaults( personSchema )

        return output

    def getUnmanagedObjectClasses( self ):
        """
            Returns a list of names of all of the object classes in 
            the LDAP schema, except for the ones already managed 
            (i.e., either in available or selected object class names).
        """
        objectClassSchema = self._schema.getAttributeSchema( 'objectClass' )
        managed = mergeLists( objectClassSchema.getValidValues(),
                              objectClassSchema.getRequiredValues() )
        unmanaged = filter( lambda i,m=managed: i not in m,
                            self._ldapObjects.keys() )
        unmanaged.sort()
        return unmanaged

    def iPlanetControllerLDIF( self, output, schema, dn, filter, objectCategory ):
        for controller in schema.getControllers():
            for controllerDN in self.user.getMatchingGroups( controller, dn ):
                output = output + """
dn: %s
changetype: modify
add: aci
aci:
 (target="ldap:///%s")
 (targetfilter=%s)
 (version 3.0; acl "%s can create/delete %s";
 allow(add, delete)
 groupdn="ldap:///%s";)
""" % ( dn, dn, filter, controller, objectCategory+'s', controllerDN )
        return output

    def iPlanetAttributeLDIF( self, output, schema, dn, filter, objectCategory,
                              role, dntype, userdn ):
        #
        # Generate the LDIF statements that describe which attributes 
        # have which permissions for which roles.
        #
        readable = schema.getReadableAttributesByRole( role )
        if len( readable ):
            output = output + """
dn: %s
changetype: modify
add: aci
aci:
 (target = "ldap:///%s")
 (targetattr = "%s")
 (targetfilter=%s)
 (version 3.0;
 acl "%s can read from %ss";
 allow (read,compare,search)
 (%s = "%s");)
""" % ( dn, dn, join(readable,' || '), filter, role, objectCategory, 
            dntype, userdn )

        writeable = schema.getWriteableAttributesByRole( role )
        if len( writeable ):
            output = output + """
dn: %s
changetype: modify
add: aci
aci:
 (target = "ldap:///%s")
 (targetattr = "%s")
 (targetfilter=%s)
 (version 3.0;
 acl "%s can write to %ss";
 allow (read,compare,search,write)
 (%s = "%s");)
""" % ( dn, dn, join(writeable,' || '), filter, role, objectCategory, 
            dntype, userdn )

        return output

    def showSchemaForm( self, message ):
        """
            Given a request for a schema management form, invoke any
            schema updates that may have been called for in the request,
            and then return the appropriate form.
        """

        #   If we got here by a form submit button, then we need to
        #   update the schema first.

        if self.request.has_key('updateSchema'):
            if not self.request.has_key('cancel'):
                message = self.updateSchemaFromRequest()
                self.aq_parent._p_changed = 1
                parent = self.aq_parent
                self = ZSchema( self.user, self.request )
                self = self.__of__( parent )

        #   Set default form, and render it.  Each form must contain a
        #   hidden tag called 'schemaFormID' whose value, with '_'
        #   prepended, and 'Form' appended, is form rendering DTMLFile
        #   call.

        if not self.request.has_key('schemaFormID'):
            self.request.set('schemaFormID','schemaObjectClasses')

        if self.request.schemaFormID == 'schemaObjectClasses':
            return self._schemaObjectClassesForm(
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaAttributes':
            return self._schemaAttributesForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaAttrsByCategory':
            return self._schemaAttrsByCategoryForm(
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaAttrsByRole':
            return self._schemaAttrsByRoleForm(
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaManaged':
            return self._schemaManagedForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaCategories':
            return self._schemaCategoriesForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaRenameCategories':
            return self._schemaRenameCategoriesForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaControllers':
            return self._schemaControllersForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaRoles':
            return self._schemaRolesForm(
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaRenameRoles':
            return self._schemaRenameRolesForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaSearching':
            return self._schemaSearchingForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'saveSettings':
            return self._saveSettingsForm( 
                        self, self.request, manage_tabs_message = message )

        else:
            message = """
                      The previous form's schemaFormID does not match a
                      known form display method.
                      """
            return self._schemaManagedForm( self, self.request,
                                            manage_tabs_message = message )

    def nextAttributeLink( self ):
        """
            Construct an href to the next managed attribute.
        """
        managed = self._schema.getManagedAttributeNames()
        ldapName = self.request.ldapName
        attributeIndex = managed.index( ldapName )

        if attributeIndex == len(managed)-1:
            return "&nbsp;"
        else:
            link = "<a href=\"%s?objectCategory=%s" % \
                   ( self.request['URL0'], self.request['objectCategory']) + \
                   "&schemaFormID=%s&ldapName=%s\">" % \
                   ( self.request['schemaFormID'], managed[ attributeIndex+1]) + \
                   "Next</a>"
            return link

    def prevAttributeLink( self):
        """
            Construct an href to the previous managed attribute.
        """
        managed = self._schema.getManagedAttributeNames()
        ldapName = self.request.ldapName
        attributeIndex = managed.index( ldapName )

        if attributeIndex == 0:
            return "&nbsp;"
        else:
            link = "<a href=\"%s?objectCategory=%s" % \
                   ( self.request['URL0'], self.request['objectCategory']) + \
                   "&schemaFormID=%s&ldapName=%s\">" % \
                   ( self.request['schemaFormID'], managed[ attributeIndex-1]) + \
                   "Previous</a>"
            return link

    def updateSchemaFromLDAP( self, user ):
        """
            Retrieves the schema from the LDAP directory server
            and updates the persisted schema collection.
        """
        # Search for the subschemaSubentry attribute of the root
        # object, which gives the DN of the schema.  Get the schema
        # objectClasses and attributesTypes, each of which is a list
        # of LDIF expressions, and parse them into dictionaries of
        # parsed objects, which have the LDIF data as attributes.

        LDAPschemaDN = user.search( "", _ldap.SCOPE_BASE, "(objectclass=*)",
            ['subschemaSubentry',])[0][1]['subschemaSubentry'][0]

        LDAPschema = user.search( LDAPschemaDN, _ldap.SCOPE_BASE,
            "(objectclass=*)", ["objectClasses","attributeTypes"] )[0][1]

        ldapObjects = {}
        for ldifExpression in LDAPschema['objectClasses']:
            parsedObject = LDIFObjectParser( ldifExpression )
            ldapObjects[lower(parsedObject.name)] = parsedObject

        #   Given the list of objectClass definitions, build up a list
        #   of attribute names available to this category of object.
        #   The required object classes are the ones expected of every
        #   object of this category in the directory.  New objects of
        #   this category will be created with objectClass set to the
        #   list of required object classes.  The set of attributes
        #   required by the required object classes will be marked as
        #   required.
        #
        #   The available object classes are those that might be
        #   present in some objects of this category in the directory.
        #   New objects of this category are not given these object
        #   class values.  The set of attributes required by the
        #   available object classes will NOT be marked as required.

        requiredLDAPAttrNames = []
        optionalLDAPAttrNames = []

        requiredObjectClasses = \
            self._schema.getAttributeSchema('objectClass').getRequiredValues()

        for objectClass in requiredObjectClasses:

            required, optional = getLDAPAttributeNames(objectClass, ldapObjects)
            requiredLDAPAttrNames = mergeLists(requiredLDAPAttrNames, required)
            optionalLDAPAttrNames = mergeLists(optionalLDAPAttrNames, optional)

        availableObjectClasses = \
            self._schema.getAttributeSchema('objectClass').getAvailableValues()

        for objectClass in availableObjectClasses:

            required, optional = getLDAPAttributeNames(objectClass, ldapObjects)
            optionalLDAPAttrNames = mergeLists(optionalLDAPAttrNames, required)
            optionalLDAPAttrNames = mergeLists(optionalLDAPAttrNames, optional)

        optionalLDAPAttrNames = filter( lambda i,r=requiredLDAPAttrNames:
                                        i not in r,
                                        optionalLDAPAttrNames )
                                  
        allowedLDAPAttrNames = mergeLists(requiredLDAPAttrNames,
                                          optionalLDAPAttrNames )

        allowedLDAPAttrNames.sort()
        self._schema._allowedLDAPAttrNames = allowedLDAPAttrNames

        # Now parse the attribute definitions.  Because attributes
        # can be subclassed from other attributes, the ldapAttributes
        # dictionary may contain parsed objects that have a specified
        # superior but no other attributes.  For attributes with
        # a superior, find the superior attribute parsed object,
        # and passing it back to the target attribute's inheritsFrom
        # method.

        ldapAttributes = {}
        for ldifExpression in LDAPschema['attributeTypes']:
            parsedObject = LDIFAttributeParser( ldifExpression )
            ldapAttributes[lower(parsedObject.name)] = parsedObject
            if parsedObject.aliases:
                for alias in parsedObject.aliases:
                    ldapAttributes[lower(alias)] = parsedObject

        for ldapAttrName in self._schema._allowedLDAPAttrNames:
            parsedAttr = ldapAttributes[lower(ldapAttrName)]
            if parsedAttr.sup:
                superiorAttrName = parsedAttr.sup
                while superiorAttrName:
                    superiorAttr = ldapAttributes[lower(superiorAttrName)]
                    parsedAttr.inheritsFrom( superiorAttr )
                    superiorAttrName = superiorAttr.sup

        #   Update the schema attribute collection.  Get the metadata
        #   from items already in the attribute collection, and update
        #   that with metadata from the LDAP schema.  New items only
        #   get LDAP schema metadata, but they are set to be unmanaged,
        #   and to be readable and writeable by the Controllers for
        #   this category of object.  Update the 'required' metadata
        #   item, and then create a new ManagedAttributeSchema object
        #   and put it in the attribute collection.  Attribute
        #   schemas no longer allowed, because the objectClass that 
        #   defined it has been dropped, are deleted.

        for ldapName in allowedLDAPAttrNames:

            if ldapName in self._schema._attributeCollection.keys():
                metadata = \
                    self._schema.getAttributeSchema(ldapName).getMetadata()
                metadata.update(
                    ldapAttributes[lower(ldapName)].getMetadata() )
            else:
                metadata = ldapAttributes[lower(ldapName)].getMetadata()
                metadata = { 'managed': FALSE,
                             'readers': self._schema._Controllers,
                             'writers': self._schema._Controllers, }

            if ldapName in requiredLDAPAttrNames:
                metadata['required'] = TRUE
                metadata['managed'] = TRUE
            else:
                metadata['required'] = FALSE

            managedAttribute = ManagedAttributeSchema( ldapName, metadata )
            self._schema._attributeCollection[ ldapName ] = managedAttribute

        for ldapName in self._schema._attributeCollection.keys():
            if not ldapName in allowedLDAPAttrNames:
                del self._schema._attributeCollection[ldapName]                

        #   Force the namingAttribute to be required (and managed).

        namingAttribute = self._schema.getNamingAttribute()
        namingAttributeAttr = self._schema.getAttributeSchema(namingAttribute)
        namingAttributeAttr.setRequired(TRUE)
        namingAttributeAttr.setManaged(TRUE)
        self._schema.setSearchKeyFriendlyNames()

        #   The list of available LDAP objects is used to check new
        #   values for the 'objectClass' attribute, so we return it
        #   for ZSchema use.  But the ManagedSchema object doesn't 
        #   need it.

        return ldapObjects
    
    def updateSchemaFromRequest( self):
        """
            Change the schema according to the REQUEST variables.
        """
        message = None

        #   Invoke a method that matches the name of the submit
        #   button (with '_' prepended).

        if self.request.has_key( 'saveAttributeSchema' ):
            message = self._saveAttributeSchema()

        elif self.request.has_key( 'setNamingAttribute' ):
            message = self._setNamingAttribute()

        elif self.request.has_key( 'moveAttributeToAvailable' ):
            message = self._moveAttributeToAvailable()

        elif self.request.has_key( 'moveAttributeToManaged' ):
            message = self._moveAttributeToManaged()

        elif self.request.has_key( 'moveObjectClassesToSelected' ):
            message = self._moveObjectClassesToSelected()

        elif self.request.has_key( 'moveObjectClassesToAvailable' ):
            message = self._moveObjectClassesToAvailable()

        elif self.request.has_key( 'dropObjectClasses' ):
            message = self._dropObjectClasses()

        elif self.request.has_key( 'addAvailableObjectClasses' ):
            message = self._addAvailableObjectClasses()

        elif self.request.has_key( 'changeDisplayCategories' ):
            message = self._changeDisplayCategories()

        elif self.request.has_key( 'renameDisplayCategories' ):
            message = self._renameDisplayCategories()

        elif self.request.has_key( 'saveAttrsByCategory' ):
            message = self._saveAttrsByCategory()

        elif self.request.has_key( 'saveAttrsByRole' ):
            message = self._saveAttrsByRole()

        elif self.request.has_key( 'addControllers' ):
            message = self._addControllers()

        elif self.request.has_key( 'deleteControllers' ):
            message = self._deleteControllers()

        elif self.request.has_key( 'changeRoles' ):
            zRoles = ZRoles( self.user, self._managedRoles )
            zRoles = zRoles.__of__( self )
            message = zRoles.updateRolesFromRequest( self.request )

        elif self.request.has_key( 'addSearchKeys' ):
            message = self._addSearchKeys()

        elif self.request.has_key( 'deleteSearchKeys' ):
            message = self._deleteSearchKeys()

        elif self.request.has_key( 'searchSave' ):
            message = self._searchSave()

        elif self.request.has_key( 'firstSearchKeys' ):
            message = self._firstSearchKeys()

        else:
            # You can get here if your form has a single text input
            # field and you press return rather than the form button.
            # Maybe there is a way to figure it out:

            message = "No method to perform requested action."

        return message

    def _addAvailableObjectClasses( self ):

        # A method for adding items to the available object class
        # list.  Each input name is checked against the LDAP
        # schema to make sure it exists.

        newClasses = self.request.get('newClasses',None)
        if newClasses is None:
            return

        schema = self._schema.getAttributeSchema( 'objectClass' )
        valid = schema.getValidValues()
        added = []
        bogus = []
        super = []

        for newClass in newClasses:
            if self._ldapObjects.has_key( lower(newClass) ):

                # Get canonical name.

                name = self._ldapObjects[lower(newClass)].name
                schema.addValidValues( name )
                valid.append( name )
                added.append( name )

                # Do any superclasses also need to be added?

                supers = getLDAPSuperClassNames( name,
                                                      self._ldapObjects )
                supersToAdd = filter( lambda i,v=valid: i not in v, supers )
                for superClass in supersToAdd:
                    name = self._ldapObjects[lower(superClass)].name
                    schema.addValidValues( name )
                    valid.append( name )
                    super.append( name )

            else:
                bogus.append( name )

        message = ""
        if len( added ) > 0:
            message = message + \
                      "Added object classes %s.  " % ( str( added ) )
        if len( super ) > 0:
            message = message + \
                      "Added missing superClasses %s.  " % ( str( super ) )
        if len( bogus ) > 0:
            message = message + \
                      "No matching LDAP classes for %s." % ( str( bogus ) )
        return message

    def _addControllers( self ):

        # A method for adding the the controllers list

        ids = self.request.get( 'newRoles' )

        zopeRoleSearchBase = \
            self.user.getLDAPDirectory().getZopeRoleSearchBase()
        for item in ids:
            if self.user.getMatchingGroups( item, zopeRoleSearchBase )!=[]:
                self._schema.addController( item )
        self._assignRolesToPermission()
        return "Added roles."

    def _addSearchKeys( self ):
        """
            Extend the list of search keys with values from the REQUEST.
        """
        self._searchSave()
        allowedKeys = self._schema.getManagedAttributeNames()
        newSearchKeys = self.request.get( 'searchKeys' )
        newSearchKeys = filter( lambda i,ids=allowedKeys: i in ids, 
                                newSearchKeys )
        self._schema.addSearchKeys( newSearchKeys )

    def _assignRolesToPermission( self ):
        """
            Sync the RoleManager permissions with the controllers list
            for this type of object.
        """
        roles = self._schema.getControllers()
        for role in roles:
            RoleManager._addRole( self.aq_parent, role )
        RoleManager.manage_permission( self.aq_parent, 
                                       self._schema._permissionName, 
                                       roles,
                                       acquire=TRUE )

    def _changeDisplayCategories( self ):

        # A method for updating the display categories list

        option = self.request.changeDisplayCategories
        ids = self.request.get( 'ids' )

        if option == " Delete " and len( ids ) > 0:
            for item in ids:
                if item in self._schema.getDisplayCategories():
                    self._schema.delDisplayCategory( item )
            message = "Deleted object categories."

        elif option == " First " and len( ids ) > 0:
            displayCategories = filter( lambda i,ids=ids: i in ids,
                                        self._schema.getDisplayCategories())+\
                                filter( lambda i,ids=ids: i not in ids,
                                        self._schema.getDisplayCategories())
            self._schema.setDisplayCategories( displayCategories )
            message = "Re-ordered display categories."

        elif option == " Add ":
            newCategory = self.request.get( 'newCategory', None )
            if newCategory:
                self._schema.addDisplayCategory( newCategory )
            message = "Added object category."

        else:
            message = ""

        return message

    def _deleteControllers( self ):

        # A method for dropping items from the controllers list

        ids = self.request.get( 'ids' )
        for item in ids:
            self._schema.delController( item )
        self._assignRolesToPermission( self.request.get('objectCategory') )
        return "Deleted roles."

    def _deleteSearchKeys( self):

        self._searchSave( )

        # Delete search keys

        ids = self.request.get( 'ids' )
        searchKeys = self._schema.getSearchKeys()
        searchKeys = filter( lambda i,ids=ids: i not in ids, searchKeys )
        self._schema.setSearchKeys( searchKeys )
        return "Deleted search keys %s." % ids

    def _dropAvailableObjectClasses( self, dropClasses ):

        # Drop the specifed object classes from the available
        # object classes list.

        objectClassSchema = self._schema.getAttributeSchema( 'objectClass' )
        objectClassSchema.delValidValues( dropClasses )
        return "Dropped specified objectClasses."

    def _dropObjectClasses( self ):

        # Drop object classes from both the available and selected
        # object class lists.  

        message = ""
        availClasses = self.request.get('availClasses',[])
        if len( availClasses ) > 0:
            message = message + \
                      self._dropSelectedObjectClasses( availClasses )

        selectClasses = self.request.get('selectClasses',[])
        if len( selectClasses ) > 0:
            message = message + \
                      self._dropAvailableObjectClasses( selectClasses )

        return message

    def _dropSelectedObjectClasses( self, dropClasses ):

        # Drop the specified classes from the selected object classes
        # list.  But you can't drop an object class subclassed by one
        # of the classes remaining in the list.  In that case, the
        # drop fails.  Compute what the new selected object classes list 
        # would be, and see if it is complete.

        objectClassSchema = self._schema.getAttributeSchema( 'objectClass' )
        requiredValues = objectClassSchema.getRequiredValues()
        for item in dropClasses:
            requiredValues.remove( item )
        if not self.checkObjectClassCompleteness( requiredValues ):

            # What classes would be stranded with a missing superclass?

            stranded = []
            for item in requiredValues:
                supers = getLDAPSuperClassNames( item,
                                                      self._ldapObjects )
                if not list1ContainsList2( requiredValues, supers ):
                    stranded.append( item )

            message = """
                      The requested object class(es), %s, could not 
                      be moved because the list contains superclasses
                      for classes %s.  Only mixin and lowest level 
                      classes can be moved.  Either move all of these
                      classes, or pick items from the second list.
                      """ % ( str(dropClasses), str(stranded) )

            return message

        #   The REQUEST is missing the 'validValues' and 'requiredValues'
        #   keys that any other attribute schema update would have.
        #   Supply them, and then call the _saveAttributeSchema method.

        objectClassSchema.setRequiredValues( requiredValues )

        message = self._saveAttributeSchema()
        return "Dropped specified objectClasses.  " + str(message)

    def _firstSearchKeys( self):

        self._searchSave()

        # Move the select keys to the head of the list

        ids = self.request.get( 'ids' )
        searchKeys = self._schema.getSearchKeys()
        searchKeys = filter(lambda i,ids=ids: i in ids, searchKeys ) \
                     + filter(lambda i,ids=ids: i not in ids, searchKeys )
        self._schema.setSearchKeys( searchKeys )
        return "Reordered search keys."

    def _moveAttributeToAvailable(self):

        # A method for making an attribute "available," rather than
        # "managed."

        selectedAttrs = self.request.get('unmanageAttrs',[])
        for item in selectedAttrs:
            self._schema.unmanageAttributeSchema( item )

        return "Updated attribute lists."

    def _moveAttributeToManaged(self):

        # A method for making an attribute "managed," rather than
        # "available."

        selectedAttrs = self.request.get('manageAttrs',[])
        for item in selectedAttrs:
            self._schema.manageAttributeSchema( item )

        return "Updated attribute lists."

    def _moveObjectClassesToAvailable(self):

        # A method for making an object class value "available,"
        # rather than "selected."  But you can't unselect an
        # object class subclassed by some other object in the
        # required list.

        availClasses = self.request.get('availClasses',[])
        message = self._dropSelectedObjectClasses(availClasses)
        if message[0:31] == "Dropped specified objectClasses.":
            # i.e., the drop worked
            objectClassSchema = self._schema.getAttributeSchema( 'objectClass' )
            objectClassSchema.addValidValues( availClasses )
            message = self._saveAttributeSchema()
            return "Moved specified objectClasses.  " + str(message)

        else:
            return message

    def _moveObjectClassesToSelected(self):

        # A method for making an object class value "selected,"
        # rather than "available."  In addition to any specified
        # classes, be sure to move their superclasses, too.

        selectClasses = self.request.get('selectClasses',[])
        supers = []
        for item in selectClasses:
            supers = mergeLists( supers,
                                 getLDAPSuperClassNames( item,
                                                         self._ldapObjects ) 
                               )
        
        objectClassSchema = self._schema.getAttributeSchema( 'objectClass' )

        requiredValues = mergeLists( objectClassSchema.getRequiredValues(),
                                     selectClasses )

        # Now reassign any superclasses that also need to move

        supersToAdd = filter( lambda i,r=requiredValues: i not in r, supers )
        if len( supersToAdd ) > 0:
            requiredValues = mergeLists( requiredValues, supersToAdd )
            message = "  Added superclasses to the required class list."

        objectClassSchema.setRequiredValues( requiredValues )

        message = self._saveAttributeSchema()

        return "Updated object class lists.  " + str(message)

    def _renameDisplayCategories(self):
        """
            Rename display categories.
        """
        #   If we got here from the _schemaRenameCategoriesForm, then 
        #   perform the rename.

        if self.request.has_key('haveNewNames'):
            ids = self.request.get( 'ids', [] )
            new_ids = self.request.get( 'new_ids', [] )
            displayCategories = self._schema.getDisplayCategories()

            # Skip empty new_ids

            while '' in new_ids:
                i = new_ids.index('')
                ids.pop(i)
                new_ids.pop(i)

            for i in range( 0, len( ids ) ):
                oldName = ids[ i ]
                newName = new_ids[ i ]

                #   Easy: change the list ...

                displayCategories[ displayCategories.index( oldName ) ] = \
                    newName

                #   Hard: change all of the attribute objects.

                for attributeSchema in self._schema._attributeCollection.values():
                    if oldName in attributeSchema.getDisplayCategories():
                        attributeSchema.delDisplayCategories( oldName )
                        attributeSchema.addDisplayCategories( newName )

            self._schema.setDisplayCategories( displayCategories )
            return "Renamed display categories"

        #   Otherwise, set REQUEST variables so that the
        #   _schemaRenameCategoriesForm gets displayed.

        else:
            self.request.set('schemaFormID', 'schemaRenameCategories')
            del self.request.form['updateSchema']

            return ""

    def _saveAttributeSchema(self):
        """
            Save an updated attribute schema.
        """
        ldapName = self.request['ldapName']
        schema = self._schema.getAttributeSchema( ldapName )

        if not ldapName == 'objectClass':

            schema.setValidValues( self.request.get('validValues') )
            schema.setRequiredValues( self.request.get('requiredValues') )

        schema.setDisplayCategories( self.request.get('displayCategories',[]) )
        schema.setFriendlyName( self.request.get('friendlyName', '') )
        schema.setManaged( self.request.get('managed', FALSE ) )
        schema.setPointsTo( self.request.get('pointsTo', '') )
        schema.setReaders( self.request.get('readers',[]) )
        schema.setRequired( self.request.get('required', FALSE ) )
        schema.setSyntaxByOID( self.request.get('syntax', None) )
        schema.setWriters( self.request.get('writers',[]) )

        return "Saved updated schema."

    def _saveAttrsByCategory(self):
        """
            Save results of the schemaAttrsByCategory form.
        """
        displayCategory = self.request[ 'displayCategory' ]
        ids = self.request['ids']

        for ldapName in self._schema.getManagedAttributeNames():

            schema = self._schema.getAttributeSchema( ldapName )
            if ldapName in ids:
                schema.addDisplayCategories( displayCategory )
            elif displayCategory in schema.getDisplayCategories():
                schema.delDisplayCategories( displayCategory )

        return """
               Saved display category \"%s\"'s presence or absence
               in each managed attribute. 
               """ % ( displayCategory )

    def _saveAttrsByRole(self):
        """
            Save results of the schemaAttrsByRole form.
        """
        roleName = self.request[ 'roleName' ]
        readIds = self.request.get('readIds',[])
        writeIds = self.request.get('writeIds',[])

        for ldapName in self._schema.getManagedAttributeNames():

            schema = self._schema.getAttributeSchema( ldapName )
            if ldapName in readIds:
                schema.addReaders( roleName )
            elif roleName in schema.getReaders():
                schema.delReaders( roleName )

            if ldapName in writeIds:
                schema.addWriters( roleName )
            elif roleName in schema.getWriters():
                schema.delWriters( roleName )

        return """
               Saved readers and writers settings for the %s role
               in each managed attribute.
               """ % ( roleName )

    def _setPermissionName( self, permissionName ):

        # Set the name of the permission that controls the creation and
        # deletion of objects of this objectCategory in LDAP.

        self._schema._permissionName = permissionName

    def _setNamingAttribute(self):

        # Set the naming attribute for this schema.

        namingAttribute = self.request.get( 'namingAttribute' )
        self._schema.setNamingAttribute( namingAttribute )

        return "Set naming attribute."

    def _searchSave(self):

        # Save search configuration specified in REQUEST

        searchFilter = self.request.get( 'searchFilter' )
        self._schema.setSearchFilter( searchFilter )

        personsOUAttr = self.request.get( 'personsOUAttr' )
        givenAttr = self.request.get( 'givenAttr' )
        if personsOUAttr == 'dn':
            self._schema.setPersonsOUAttr( 'dn' )
        elif personsOUAttr == 'useGivenAttr':
            self._schema.setPersonsOUAttr( givenAttr )

        searchSortKey = self.request.get( 'searchSortKey' )
        if searchSortKey in self._schema.getSearchKeys():
            self._schema.setSearchSortKey( searchSortKey )
        else:
            self._schema.setSearchSortKey(self._schema.getSearchKeys()[0])

        for key in self._schema.getSearchKeys():
            keyName = self.request.get( key+'_name', None )
            if keyName:
                self._schema.setSearchKeyName( key, keyName )

        return "Saved changes."

####################################################################
#
#   The method below is an auxiliary method that would be marked
#   as private class methods in other OO languages.  It does not
#   need any instance specific information and it should not be
#   "imported" into any other python module.
#
####################################################################

def getLDAPAttributeNames(objectClass, objectSchemas):
    """
        Returns lists of required and optional attributes for the
        given LDAP object class.
    """
    if not objectSchemas.has_key( lower(objectClass) ):
        return [],[]
    LDIFObject = objectSchemas[lower(objectClass)]
    required = LDIFObject.mustContain
    optional = LDIFObject.mayContain

    if LDIFObject.superClass:
        supRequired, supOptional = \
            getLDAPAttributeNames(LDIFObject.superClass, objectSchemas)
        required = mergeLists( required, supRequired )
        optional = mergeLists( optional, supOptional )

    return required, optional

def getLDAPSuperClassNames( objectClass, objectSchemas ):
    """
        Returns a list of superclass names for the given LDAP object
        class.
    """
    LDIFObject = objectSchemas[lower(objectClass)]

    if LDIFObject.superClass:
        super = [ LDIFObject.superClass, ]
        super.extend( getLDAPSuperClassNames( LDIFObject.superClass,
                                              objectSchemas ) )
        return super

    else:
        return []


